from smokesim.smoke import SmokeMachine
from smokesim.engine import EngineTypes, Engine

import pygame
from typing import Optional, Tuple
from pathlib import Path
import numpy as np
import cv2


class Augmentation:
    def __init__(
        self,
        image_path: Optional[Path] = Path("assets/me.jpg"),
        screen_dim: Tuple[int, int] = (500, 700),
        smoke_machine: Optional[SmokeMachine] = None,
        random_seed: int = 100,
        engine_type: EngineTypes = EngineTypes.PYGAME,
    ):
        """
        Initialize the Augmentation class.

        Args:
        - image_path (Optional[Path], optional): The path to the image. Defaults to Path('assets/me.jpg').
        - screen_dim (Tuple[int, int], optional): The screen dimensions. Defaults to (500, 700).
        - smoke_machine (Optional[SmokeMachine], optional): The smoke machine. Defaults to None.
        - random_seed (int, optional): The random seed. Defaults to 100.
        - engine_type (EngineTypes, optional): The engine type. Defaults to EngineTypes.PYGAME.

        """

        self.image_path = image_path
        self.image = None
        self.screen_dim = screen_dim
        self.engine_type = engine_type
        self.engine = Engine(engine_type)
        self.read_image(image_path)

        self.screen = smoke_machine.screen if smoke_machine else self.make_screen()
        self.smoke_machine = (
            smoke_machine
            if smoke_machine
            else SmokeMachine(self.screen, auto_draw=False, random_seed=random_seed)
        )
        self.smoke_machine.random_seed = random_seed
        self.display_image()

    def display_image(self):
        """
        A method to display the image on the screen.
        """
        if self.screen_dim == self.image.get_size():
            self.screen.blit(self.image, (0, 0))
        else:
            self.image = pygame.transform.scale(self.image, self.screen_dim)
        self.screen.blit(self.image, (0, 0))
        return self

    def read_image(self, image_path: Path) -> pygame.Surface:
        """
        A method to read an image.

        Args:
        - image_path (Path): The path to the image.

        Returns:
        - pygame.Surface: The image.
        """

        if image_path is not None:
            self.image = pygame.image.load(str(image_path))
            if not image_path.exists():
                Warning("No image path provided. Creating a blank image.")
                image_path = None
        if image_path is None:
            self.image = np.zeros((self.screen_dim[0], self.screen_dim[1], 3))

            self.image = pygame.surfarray.make_surface(self.image)

            self.image.set_alpha(255)

        self.image = pygame.transform.scale(self.image, self.screen_dim)
        blank_image = pygame.surfarray.make_surface(
            np.zeros((self.screen_dim[0], self.screen_dim[1], 3))
        )
        blank_image.set_alpha(255)
        self.blank_image = pygame.transform.scale(blank_image, self.screen_dim)
        return self

    def make_screen(self) -> pygame.Surface:
        """
        A method to make a screen.

        Returns:
        - pygame.Surface: The screen.
        """
        pygame.init()
        self.clock = pygame.time.Clock()
        screen = pygame.display.set_mode(self.screen_dim, flags=pygame.HIDDEN)
        return screen

    def add_smoke(self, args: dict):
        """
        A method to add smoke to the screen.

        Args:
        - args (dict): The arguments to pass to the smoke machine.
        """
        self.smoke_machine.add_smoke(args)

    def augment_iter(
        self,
        steps: int,
        time_step: float,
        image: Optional[np.ndarray] = None,
        history_path: Optional[Path] = None,
    ):
        """
        A method to augment the image with smoke.

        Args:
        - steps (int): The number of steps to augment the image.
        - time_step (float): The time step to augment the image.
        - image (Optional[np.ndarray]): The image to augment.
        - history_path (Path): The path to save the history.

        Yields:
        - np.ndarray: The augmented image.
        """
        if image is not None:
            self.image = pygame.surfarray.make_surface(image)
            self.image = pygame.transform.scale(self.image, self.screen_dim)

        self.writer = None
        if history_path:
            self.writer = cv2.VideoWriter(
                str(history_path),
                cv2.VideoWriter_fourcc(*"mp4v"),
                30,
                self.image.get_size(),
            )
        for t in range(steps):
            # print(t)
            self.screen.blit(self.image, (0, 0))
            self.smoke_machine.update(time_step=time_step)
            self.smoke_machine.draw(self.screen)

            pygame.display.flip()
            rgb_array = pygame.surfarray.array3d(pygame.display.get_surface())

            self.screen.blit(self.blank_image, (0, 0))
            self.smoke_machine.draw(self.screen)

            pygame.display.flip()
            rgb_mask_array = pygame.surfarray.array3d(pygame.display.get_surface())
            rotated_rgb_mask_array = cv2.rotate(rgb_mask_array, cv2.ROTATE_90_CLOCKWISE)
            if self.writer:
                self.writer.write(
                    cv2.cvtColor(
                        rotated_rgb_mask_array,
                        cv2.COLOR_RGB2BGR,
                    )
                )

            yield cv2.rotate(rgb_array, cv2.ROTATE_90_CLOCKWISE), rotated_rgb_mask_array

    def augment(
        self,
        steps: int = 2,
        time_step: float = 30,
        image: Optional[np.ndarray] = None,
        history_path: Optional[Path] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        A method to augment the image with smoke.

        Args:
        ----------------
        - steps (int, optional): The number of steps to augment the image. Defaults to 2.
        - time_step (float, optional): The time step to augment the image. Defaults to 30.
        - image (Optional[np.ndarray], optional): The image to augment. Defaults to None.

        Returns:
        ----------------
        - np.ndarray: The smoke overlayed image.
        - np.ndarray: The smoke only image.
        """
        if image is not None:
            self.image = pygame.surfarray.make_surface(image)
            self.image = pygame.transform.scale(self.image, self.screen_dim)
        for rgb_array, rgb_mask_array in self.augment_iter(
            steps, time_step, image, history_path=history_path
        ):
            pass
        if self.writer:
            self.writer.release()
        self.last_aug_img, self.last_aug_mask = rgb_array, rgb_mask_array
        return self.last_aug_img, self.last_aug_mask

    def save_as(self, out_dir: Path = Path("assets/augmented_smoke.png")):
        """
        A method to save the augmented image.

        Args:
        - out_dir (Path, optional): The output directory to save the image. Defaults to Path('assets/augmented_smoke.png').

        """
        # pygame.image.save(self.screen, str(out_dir))
        cv2.imwrite(str(out_dir), cv2.cvtColor(self.last_aug_img, cv2.COLOR_RGB2BGR))
        cv2.imwrite(
            str(out_dir.parent / (out_dir.stem + "_mask.png")),
            cv2.cvtColor(self.last_aug_mask, cv2.COLOR_RGB2BGR),
        )
        print(f"Saved augmented image as {out_dir}")

    def end(self):
        """
        A method to end the pygame session.
        """
        pygame.quit()



from smokesim.base import BaseSim, BaseProperty
from smokesim.constants import CLOUD_MASK

import pygame
import cv2
import numpy as np
from typing import Optional, Tuple


class Sprite(BaseProperty):
    """
    A dataclass to represent a sprite.
    """

    width: int = 20
    height: int = 20
    opacity_mask: np.ndarray = CLOUD_MASK
    color: tuple = (255, 255, 255)

    class Config:
        arbitrary_types_allowed = True


class ParticleProperty(BaseProperty):
    """
    A dataclass to represent the arguments for a particle.


    - vx (Optional[float], optional): The x velocity of the particle. Defaults to None.
    - startvy (Optional[float], optional): The y velocity of the particle. Defaults to None.
    - scale (Optional[float], optional): The scale of the particle. Defaults to None.
    - lifetime (Optional[int], optional): The lifetime of the particle. Defaults to None.
    - age (int, optional): The age of the particle. Defaults to 0.
    - min_vx (float, optional): The minimum x velocity of the particle. Defaults to -4/100.
    - max_vx (float, optional): The maximum x velocity of the particle. Defaults to 4/100.
    - min_vy (float, optional): The minimum y velocity of the particle. Defaults to -4/10.
    - max_vy (float, optional): The maximum y velocity of the particle. Defaults to -1/10.
    - min_scale (int, optional): The minimum scale of the particle. Defaults to 20.
    - max_scale (int, optional): The maximum scale of the particle. Defaults to 40.
    - min_lifetime (float, optional): The minimum lifetime of the particle. Defaults to 2000.
    - max_lifetime (float, optional): The maximum lifetime of the particle. Defaults to 8000.
    - color (Tuple[float, float, float], optional): The color of the particle. Defaults to (167, 167, 167).
    - smoke_sprite_size (int, optional): The size of the sprite. Defaults to 20.
    - fade_speed (int, optional): The speed at which the particle fades. Defaults to 1.
    - alpha (int, optional): The alpha of the particle. Defaults to 255.
    """

    startvx: Optional[float] = None
    startvy: Optional[float] = None
    scale: Optional[float] = None
    lifetime: Optional[int] = None
    age: int = 0
    min_vx: float = -4 / 100
    max_vx: float = 4 / 100
    min_vy: float = -4 / 10
    max_vy: float = -1 / 10
    # select random scale between min_scale and max_scale
    min_scale: int = 20
    max_scale: int = 40
    # select a final scale between 0.1 and 1.5 times the initial scale
    scale_range: Tuple[int, int] = (0.01, 1.5)
    min_lifetime: float = 2000
    max_lifetime: float = 8000
    color: Tuple[float, float, float] = (24, 46, 48)
    smoke_sprite_size: int = 20
    fade_speed: int = 1
    alpha: int = 255
    # seed:int = 100
    id: str = ""


class Particle(BaseSim):
    def __init__(self, x: int, y: int, property: ParticleProperty):
        """
        A class to represent a particle. Particles are used to create smoke.

        Args:
         - x (int): The x coordinate of the particle.
        - y (int): The y coordinate of the particle.
        - property (ParticleProperty): The properties of the particle.
            - vx (Optional[float], optional): The x velocity of the particle. Defaults to None.
            - startvy (Optional[float], optional): The y velocity of the particle. Defaults to None.
            - scale (Optional[float], optional): The scale of the particle. Defaults to None.
            - lifetime (Optional[int], optional): The lifetime of the particle. Defaults to None.
            - age (int, optional): The age of the particle. Defaults to 0.
            - min_vx (float, optional): The minimum x velocity of the particle. Defaults to -4/100.
            - max_vx (float, optional): The maximum x velocity of the particle. Defaults to 4/100.
            - min_vy (float, optional): The minimum y velocity of the particle. Defaults to -4/10.
            - max_vy (float, optional): The maximum y velocity of the particle. Defaults to -1/10.
            - min_scale (int, optional): The minimum scale of the particle. Defaults to 20.
            - max_scale (int, optional): The maximum scale of the particle. Defaults to 40.
            - min_lifetime (float, optional): The minimum lifetime of the particle. Defaults to 2000.
            - max_lifetime (float, optional): The maximum lifetime of the particle. Defaults to 8000.
            - color (Tuple[float, float, float], optional): The color of the particle. Defaults to (167, 167, 167).
            - smoke_sprite_size (int, optional): The size of the sprite. Defaults to 20.
            - fade_speed (int, optional): The speed at which the particle fades. Defaults to 1.
            - alpha (int, optional): The alpha of the particle. Defaults to 255.
        """
        super().__init__(property)

        # self.seed = property.seed
        # self.float_in_range = lambda start, end, random_state: start + np.random.random() * (end - start)
        self.property = property
        self.x = x
        self.y = y
        self.startvx = (
            property.startvx
            if property.startvx is not None
            else self.float_in_range(property.min_vx, property.max_vx)
        )
        self.startvy = (
            property.startvy
            if property.startvy is not None
            else self.float_in_range(property.min_vy, property.max_vy)
        )
        self.scale = (
            property.smoke_sprite_size
            if property.scale is not None
            else int(self.float_in_range(property.min_scale, property.max_scale))
        )
        self.lifetime = (
            property.lifetime
            if property.lifetime is not None
            else self.float_in_range(property.min_lifetime, property.max_lifetime)
        )
        self.age = property.age
        self.color = property.color
        self.smoke_sprite_size = property.smoke_sprite_size
        self.final_scale = self.float_in_range(
            self.scale * property.scale_range[0], self.scale * property.scale_range[1]
        )
        self.scale_step = (self.final_scale - self.scale) / self.lifetime
        self.vy = self.startvy
        self.vx = self.startvx
        self.alpha = property.alpha
        self.fade_speed = property.fade_speed
        self.is_alive = True
        self.sprite_paint = self.make_sprite()

    @property
    def position(self):
        return (self.x, self.y)

    def paint_sprite(self, sprite: Sprite) -> pygame.Surface:
        """
        A method to paint a sprite.

        Args:
        - sprite (Sprite): The sprite to paint.

        Returns:
        - pygame.Surface: The painted sprite.
        """
        surface = pygame.Surface((sprite.width, sprite.height), pygame.SRCALPHA)
        pixels = pygame.PixelArray(surface)
        opacities = cv2.resize(
            sprite.opacity_mask,
            (sprite.width, sprite.height),
            interpolation=cv2.INTER_NEAREST,
        )
        for x in range(sprite.width):
            for y in range(sprite.height):
                pixels[x, y] = (*sprite.color, opacities[x, y])
        del pixels
        # surface = pygame.transform.smoothscale(
        #     surface, (sprite.width // 2, sprite.height // 2)
        # )
        surface = pygame.transform.smoothscale(surface, (sprite.width, sprite.height))
        self.surface = surface
        return surface

    def make_sprite(self) -> pygame.Surface:
        """
        A method to make a sprite.

        Returns:
        - pygame.Surface: The sprite.
        """
        self.sprite = Sprite(
            color=self.color,
            width=self.scale,
            height=self.scale,
        )
        self.sprite_paint = self.paint_sprite(self.sprite)
        return self.sprite_paint

    def update(self, time_step: float = 1):
        """
        A method to update the particle.

        Args:
        - time_step (float, optional): The time_step to update the particle by. Defaults to 1.

        """
        self.age += time_step
        self.x += self.vx * time_step
        self.y += self.vy * time_step
        frac = self.age / self.lifetime
        frac = frac**0.5
        self.vy = (1 - frac) * self.startvy
        self.vx = (1 - frac) * self.startvx
        scale_frac = self.age * self.scale_step

        height = self.scale + scale_frac
        width = self.scale + scale_frac
        self.alpha -= self.fade_speed

        if (
            (self.alpha < 0)
            or (self.age > self.lifetime)
            or (self.scale < 1)
            or (height < 0)
            or (width < 0)
        ):
            self.is_alive = False
            return None
        # if self.scale > :
        # self.sprite.width = int(self.smoke_sprite_size * self.scale)
        # self.sprite.height = int(self.smoke_sprite_size * self.scale)
        # if self.sprite.width < 1 or self.sprite.height < 1:
        #     self.is_alive = False

        # surface = pygame.transform.smoothscale(
        #     self.surface, (self.sprite.width // 2, self.sprite.height // 2)
        # )
        surface = pygame.transform.smoothscale(self.surface, (height, width))
        self.sprite_paint = surface

    def draw(self, screen):
        """
        A method to draw the particle.

        Args:
        - screen (pygame.Surface): The screen to draw the particle on.
        """
        if self.is_alive:
            self.sprite_paint.set_alpha(self.alpha)
            screen.blit(self.sprite_paint, (int(self.x), int(self.y)))

        if (
            self.x < 0
            or self.y < 0
            or self.x > screen.get_width()
            or self.y > screen.get_height()
        ):
            self.is_alive = False

    def __del__(self):
        try:
            del self.sprite_paint
            del self.sprite
            del self.surface
        except AttributeError:
            pass



from typing import List, Optional, Tuple

import pygame

from smokesim.base import BaseProperty, BaseSim
from smokesim.particle import Particle, ParticleProperty


class SmokeProperty(BaseProperty):
    """
    A dataclass to represent the arguments for a smoke.

    - origin (Tuple[int, int], optional): The origin of the smoke. Defaults to (100, 100).
    - particle_count (int, optional): The number of particles in the smoke. Defaults to 100.
    - color (Tuple[int, int, int], optional): The color of the smoke. Defaults to (167, 167, 167).
    - particle_args (dict, optional): Arguments to pass to the Particle class. Defaults to {}.
    - sprite_size (int, optional): The size of the sprite. Defaults to 20.
    - lifetime (int, optional): The lifetime of the smoke. Defaults to -1.
    - age (int, optional): The age of the smoke. Defaults to 0.
    - id (int, optional): The id of the smoke. Defaults to 0.
    """

    origin: Tuple[int, int] = (100, 100)
    particle_count: int = 100
    color: Tuple[int, int, int] = (24, 46, 48)
    particle_property: Optional[ParticleProperty] = ParticleProperty()
    sprite_size: int = 20
    lifetime: int = -1
    age: int = 0
    id: int = 0


class Smoke(BaseSim):
    def __init__(self, screen: pygame.Surface, smoke_property: SmokeProperty):
        """
        Class to represent a smoke object. Smoke is made up of particles.

        Args:
        - screen (pygame.Surface): The screen to draw the smoke on.
        - smoke_property (SmokeProperty): The properties of the smoke.

        """
        super().__init__(smoke_property)
        self.origin = smoke_property.origin
        self.id = smoke_property.id
        self.screen = screen
        self.particle_count = smoke_property.particle_count
        self.color = smoke_property.color
        self.sprite_size = smoke_property.sprite_size
        self.lifetime = smoke_property.lifetime
        self.age = smoke_property.age
        self.particles: List[Particle] = []
        self.particles_until_now = 0
        self.particle_property = smoke_property.particle_property
        self.create_particles(self.particle_property)

    def create_particles(self, particle_property: Optional[ParticleProperty] = None):
        """
        A method to create particles for the smoke.

        Args:
        - particle_property (Optional[ParticleProperty], optional): The properties of the particles. Defaults to None.
        """
        particles = []
        for p in range(self.particle_count):
            particle_id = f"{self.id}_{self.particles_until_now}"
            x, y = self.origin
            if particle_property:
                particle_property.random_seed = self.random_state.randint(0, 100000)
            else:
                particle_property = ParticleProperty(
                    color=self.color, smoke_sprite_size=self.sprite_size
                )
                particle_property.random_seed = self.random_state.randint(0, 100000)

            particle_property.id = particle_id
            particle = Particle(x, y, particle_property)

            particles.append(particle)
            self.particles_until_now += 1
            # print(p, particle_id)
        self.particles.extend(particles)

    def update(self, time_step: float = 30):
        """
        A method to update the smoke.

        Args:
        - time_step (float, optional): The time_step to update the smoke by. Defaults to 30.
        """
        self.age += time_step
        new_particles = []
        for particle in self.particles:
            particle.update(time_step)
            if particle.is_alive:
                # particle.draw(self.screen)
                new_particles.append(particle)
            else:
                del particle
        if self.lifetime > 0 and self.age > self.lifetime:
            for p in self.particles:
                del p
            self.particles = []
        else:
            self.create_particles(self.particle_property)
        # print(f"ID: {self.id}, Num particles: {len(self.particles)}")

    def draw(self, screen: Optional[pygame.Surface] = None):
        screen = screen if screen else self.screen
        for particle in self.particles:
            particle.draw(screen)


class SmokeMachine:
    def __init__(
        self,
        screen: pygame.Surface,
        default_particle_count: int = 100,
        default_color: Tuple[int, int, int] = (24, 46, 48),
        default_sprite_size: int = 20,
        random_seed: int = 100,
        versbose: bool = False,
        auto_draw: bool = True,
    ):
        """
        A class to represent a smoke machine. The smoke machine creates and manages smoke objects.

        Args:
        - screen (pygame.Surface): The screen to draw the smoke on.
        - default_particle_count (int, optional): The default number of particles in the smoke. Defaults to 100.
        - default_color (Tuple[int, int, int], optional): The default color of the smoke. Defaults to (167, 167, 167).
        - default_sprite_size (int, optional): The default size of the sprite. Defaults to 20.

        """
        self.color = default_color
        self.sprite_size = default_sprite_size
        self.particle_count = default_particle_count
        self.screen = screen
        self.time_step = 0
        self.smokes: List[Smoke] = []
        self.last_smoke_id = -1
        self.random_seed = random_seed
        self.verbose = versbose
        self.auto_draw = auto_draw

    def add_smoke(self, args: dict):
        """
        A method to add smoke to the smoke machine.

        Args:
        - args (dict): The arguments to pass to the Smoke class.
            If `particle_args` is in the args, it will be passed to the ParticleProperty class.

        """

        if "color" not in args:
            args["color"] = self.color
        if "particle_count" not in args:
            args["particle_count"] = self.particle_count
        if "sprite_size" not in args:
            args["sprite_size"] = self.sprite_size
        if "id" not in args:
            args["id"] = self.last_smoke_id + 1
        if "random_seed" not in args:
            args["random_seed"] = self.random_seed
        if "particle_args" in args:
            particle_args = args["particle_args"]
            if particle_args.get("random_seed") is None:
                particle_args["random_seed"] = self.random_seed
            particle_property = ParticleProperty(**particle_args)
            # print(f"Particle property: {particle_property}")
            args["particle_property"] = particle_property
            del args["particle_args"]
        else:
            args["particle_property"] = ParticleProperty(random_seed=self.random_seed)
        smoke_property = SmokeProperty(**args)
        smoke = Smoke(self.screen, smoke_property)
        self.smokes.append(smoke)
        self.last_smoke_id += 1
        if self.verbose:
            print(f"Added smoke with id: {smoke.id}, particles: {len(smoke.particles)}")

    def empty(self):
        """
        A method to empty the smoke machine.
        """
        print("Emptying smoke")
        for s in self.smokes:
            s.age = s.lifetime
            for p in s.particles:
                p.age = p.lifetime

                del p
            del s
        self.smokes = []

    def update(self, time_step: float = 30):
        """
        A method to update the smoke machine.

        Args:
        - time_step (float, optional): The time_step to update the smoke machine by. Defaults to 30.
        """
        self.time_step += time_step
        new_smokes = []
        for smoke in self.smokes:
            smoke.update(time_step)
            if smoke.age > smoke.lifetime and smoke.lifetime > 0:
                for p in smoke.particles:
                    p.age = p.lifetime
                    del p
                del smoke
            else:
                new_smokes.append(smoke)
        self.smokes = new_smokes

        if self.auto_draw:
            self.draw()

    def draw(self, screen: Optional[pygame.Surface] = None):
        """
        A method to draw the smoke machine.

        Args:
        - screen (Optional[pygame.Surface], optional): The screen to draw the smoke machine on. Defaults to None.
        """
        screen = screen if screen else self.screen
        for smoke in self.smokes:
            smoke.draw(screen)



# the idea here is that I should be able to select an engine to draw the particles on a screen
# the screen could be a pygame screen or a PIL image
# an eninge could be pygame, or PIL
# hope is that the results from each engine will be the same

from enum import Enum
from typing import Optional, Tuple
from pathlib import Path
import numpy as np


class EngineTypes(Enum):
    PYGAME = "pygame"
    PIL = "pil"


class BaseEngine:
    def __init__(self, enigne_type: EngineTypes = EngineTypes.PYGAME):
        self.engine_type = enigne_type

    def get_screen(self):
        pass


class Engine(BaseEngine):
    def __init__(self, engine_type: EngineTypes = EngineTypes.PYGAME):
        super().__init__(engine_type)
        if engine_type == EngineTypes.PYGAME:
            import pygame

            self.engine = pygame
            self.screen: Optional[pygame.Surface] = None
        elif engine_type == EngineTypes.PIL:
            from PIL import Image

            self.engine = Image
            self.screen: Optional[Image.Image] = None

    def blit(self, screen, image, pos):
        if self.engine_type == EngineTypes.PYGAME:
            screen.blit(image, pos)
        elif self.engine_type == EngineTypes.PIL:
            screen.paste(image, pos)

    def display_image(self):
        if self.engine_type == EngineTypes.PYGAME:
            if self.screen_dim == self.image.get_size():
                self.screen.blit(self.image, (0, 0))
            else:
                self.image = self.engine.transform.scale(self.image, self.screen_dim)
            self.screen.blit(self.image, (0, 0))

        if self.engine_type == EngineTypes.PIL:
            if self.screen_dim == self.image.size:
                self.screen.paste(self.image, (0, 0))
            else:
                self.image = self.image.resize(self.screen_dim)
            self.screen.paste(self.image, (0, 0))

    def make_screen(self, screen_dim: Tuple[int, int] = (500, 700)):
        self.screen_dim = screen_dim
        if self.engine_type == EngineTypes.PYGAME:
            self.engine.init()
            screen = self.engine.display.set_mode(
                self.screen_dim, flags=self.engine.HIDDEN
            )
            return screen
        elif self.engine_type == EngineTypes.PIL:

            return self.engine.new("RGB", (500, 700), (0, 0, 0))

    def make_surface(self, image: np.ndarray):
        if self.engine_type == EngineTypes.PYGAME:
            self.image = self.engine.surfarray.make_surface(image)
            self.image = self.engine.transform.scale(self.image, self.screen_dim)
        elif self.engine_type == EngineTypes.PIL:
            self.image = self.engine.fromarray(image)
            self.image = self.image.resize(self.screen_dim)
        return self.image

    def read_image(self, image_path: Optional[Path] = None):
        """
        A method to read an image.

        Args:
        - image_path (Path): The path to the image.

        """
        if self.engine_type == EngineTypes.PYGAME:
            if image_path is not None:
                self.image = self.engine.image.load(str(image_path))
                if not image_path.exists():
                    Warning("No image path provided. Creating a blank image.")
                    image_path = None
            if image_path is None:
                self.image = np.zeros((self.screen_dim[0], self.screen_dim[1], 3))

                self.image = self.engine.surfarray.make_surface(self.image)

            self.image.set_alpha(255)

            self.image = self.engine.transform.scale(self.image, self.screen_dim)
            blank_image = self.engine.surfarray.make_surface(
                np.zeros((self.screen_dim[0], self.screen_dim[1], 3))
            )
            blank_image.set_alpha(255)
            self.blank_image = self.engine.transform.scale(blank_image, self.screen_dim)

        elif self.engine_type == EngineTypes.PIL:
            if image_path is not None:
                self.image = self.engine.open(image_path)
                if not image_path.exists():
                    Warning("No image path provided. Creating a blank image.")
                    image_path = None
            if image_path is None:
                self.image = np.zeros((self.screen_dim[0], self.screen_dim[1], 3))
                self.image = self.engine.fromarray(self.image)
            self.image.set_alpha(255)
            self.image = self.image.resize(self.screen_dim)
            blank_image = self.engine.new("RGB", (500, 700), (0, 0, 0))
            self.blank_image = blank_image.resize(self.screen_dim)

    def end(self):
        if self.engine_type == EngineTypes.PYGAME:
            self.engine.quit()
        elif self.engine_type == EngineTypes.PIL:
            pass



"""
A basic visualization module for plotting images and subplots.
I first added it here: https://github.com/Crazz-Zaac/computer-vision-experiments/blob/main/cv_expt/vis/visualization.py
"""

import numpy as np
from matplotlib import pyplot as plt
from typing import Tuple, List, Optional


def show_image(
    image: np.ndarray,
    title: Optional[str] = None,
    fig_size: Tuple[int, int] = (10, 10),
    axis: bool = False,
    show: bool = False,
    cmap: str = "gray",
):
    """
    A function to plot an image.

    Args:
    image: Image to plot.
    title: Title of the image.
    fig_size: Size of the figure.
    cmap: Colormap for the image.
    axis: Whether to show axis or not.

    Returns:
    fig: Figure object.

    """
    fig, ax = plt.subplots(figsize=fig_size)
    ax.imshow(image, vmax=255, vmin=0, cmap=cmap)
    if title:
        ax.set_title(title)
    if not axis:
        ax.axis("off")
    if show:
        plt.show()
    else:
        return fig


def subplot_images(
    image: List[np.ndarray],
    titles: Optional[List[str]] = None,
    fig_size: Tuple[int, int] = (10, 10),
    order: Tuple[int, int] = (1, 1),
    axis: bool = False,
    show: bool = False,
):
    """
    A function to plot multiple images in a subplot.

    Args:
    image: List of images to plot.
    titles: List of titles for each image.
    fig_size: Size of the figure.
    cmap: Colormap for the images.
    order: Tuple of number of rows and columns for the subplot.
    axis: Whether to show axis or not.

    Returns:
    fig: Figure object.

    """

    if len(image) == 1:
        return show_image(
            image[0],
            titles[0] if titles else None,
            fig_size if fig_size else (5, 5),
            axis,
        )
    elif len(image) == 0:
        Warning("No images to plot.")
        return None

    order = (
        (order[0], len(image) // order[0])
        if order[1] == -1
        else (len(image) // order[1], order[1]) if order[0] == -1 else order
    )

    fig, axs = plt.subplots(order[0], order[1], figsize=fig_size)
    if order[0] == 1 and order[1] == 1:
        axs = np.array([axs])
    for i, ax in enumerate(axs.flatten()):
        ax.imshow(image[i], vmax=255, vmin=0)
        if titles:
            ax.set_title(titles[i])
        if not axis:
            ax.axis("off")
    if show:
        plt.show()
    else:
        return fig



from pydantic import BaseModel, ConfigDict
import numpy as np


class BaseProperty(BaseModel):
    """
    Just to make sure same random states will be used everywhere.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    random_seed: int = 1000


class BaseSim:
    """
    Base class for simulation
    """

    def __init__(self, property: BaseProperty):
        self.property = property
        self.base_random_state = np.random.RandomState(property.random_seed)

    @property
    def random_state(self):
        # return np.random.RandomState(self.base_random_state.randint(0, 1000000))
        return self.base_random_state

    def float_in_range(self, start, end):
        return start + self.random_state.random() * (end - start)

    @staticmethod
    def update(self, **kwargs):
        raise NotImplementedError



from setuptools import setup

# setup for gradls
setup(
    name="smokesim",
    version="0.0.21",
    packages=["smokesim"],
    install_requires=[
        "numpy",
        "pygame==2.5.2",
        "pygame_gui==0.6.9",
        # "mediapipe==0.10.14",
        "black==24.4.2",
        "pydantic==2.7.1",
        "opencv-python<4.10.0",
    ],
    author="Ramkrishna Acharya(QViper)",
    author_email="qramkrishna@gmail.com",
    description="A package for doing smoke simulation.",
    long_description=open("Readme.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/q-viper/SmokeSim",
    keywords=["smoke", "simulation"],
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    download_url="https://github.com/q-viper/SmokeSim/archive/refs/tags/v0.0.2.tar.gz",
    homepage="https://q-viper.github.io/SmokeSim/",
)
